import numpy as np
import tensorflow as tf
import pandas as pd
import geopandas as gpd
import matplotlib.pyplot as plt
from shapely.geometry import LineString
from keras.layers.core import Dense, Activation, Dropout
from keras.layers import LSTM
from keras.models import Sequential, load_model
from keras.callbacks import CSVLogger, ReduceLROnPlateau
from keras.optimizers import adam_v2
import warnings

warnings.filterwarnings("ignore")
np.random.seed(120)
tf.random.set_seed(120)
# 支持中文
plt.rcParams['font.sans-serif'] = ['SimHei']  # 用来正常显示中文标签
plt.rcParams['axes.unicode_minus'] = False  # 用来正常显示负号

import math


def direction(lon1, lat1, lon2, lat2):
    if lon1 < lon2:
        if lat1 < lat2:
            # 第一象限
            return math.atan((lon2 - lon1) / (lat2 - lat1)) * 180 / math.pi
        elif lat1 >= lat2:
            # 第二象限
            return 90 + math.atan((lat1 - lat2) / (lon2 - lon1)) * 180 / math.pi
    elif lon1 > lon2:
        if lat1 <= lat2:
            return 270 + math.atan((lat2 - lat1) / (lon1 - lon2)) * 180 / math.pi
        elif lat1 > lat2:
            return 180 + math.atan((lon1 - lon2) / (lat1 - lat2)) * 180 / math.pi
    else:
        if lat1 < lat2:
            return 0
        else:
            return 180


def createSequence(data, window=10, maxmin=None, angle1=0, angle2=90, types=False):
    train_seq = []
    train_label = []
    m = 0
    n = 0
    if maxmin is None:
        nor = np.array(data.loc[:, ['lon', 'lat', 'sog', 'cog']])
        m = nor.max(axis=0)
        n = nor.min(axis=0)
    else:
        m, n = maxmin
    for traj_id in set(data['traj_id']):
        data_temp = data.loc[data.traj_id == traj_id]
        first_lon = data_temp.loc[0, 'lon']
        first_lat = data_temp.loc[0, 'lat']
        end_lon = data_temp.loc[data_temp.shape[0] - 1, 'lon']
        end_lat = data_temp.loc[data_temp.shape[0] - 1, 'lat']

        ang = direction(first_lon, first_lat, end_lon, end_lat)
        # 计算角度
        if types is True:
            if (ang > angle1) and (ang < angle2):
                continue
        else:
            if (ang < angle1) or (ang > angle2):
                continue

        data_temp = np.array(data_temp.loc[:, ['lon', 'lat', 'sog', 'cog']])
        # 标准化
        data_temp = (data_temp - n) / (m - n)

        for i in range(data_temp.shape[0] - window):
            x = []
            for j in range(i, i + window):
                x.append(list(data_temp[j, :]))
            train_seq.append(x)
            train_label.append(data_temp[i + window, :])

    train_seq = np.array(train_seq, dtype='float64')
    train_label = np.array(train_label, dtype='float64')

    return train_seq, train_label, [m, n]


# 多维反归一化
def FNormalizeMult(y_pre, y_true, max_min):
    [m1, n1, s1, c1], [m2, n2, s2, c2] = max_min
    # [m1,n1],[m2,n2] = max_min
    y_pre[:, 0] = y_pre[:, 0] * (m1 - m2) + m2
    y_pre[:, 1] = y_pre[:, 1] * (n1 - n2) + n2
    y_pre[:, 2] = y_pre[:, 2] * (s1 - s2) + s2
    y_pre[:, 3] = y_pre[:, 3] * (c1 - c2) + c2
    y_true[:, 0] = y_true[:, 0] * (m1 - m2) + m2
    y_true[:, 1] = y_true[:, 1] * (n1 - n2) + n2
    y_true[:, 2] = y_true[:, 2] * (s1 - s2) + s2
    y_true[:, 3] = y_true[:, 3] * (c1 - c2) + c2

    # 计算距离
    y_pre = np.insert(y_pre, y_pre.shape[1],
                      get_distance_hav(y_true[:, 1], y_true[:, 0], y_pre[:, 1], y_pre[:, 0]), axis=1)

    return y_pre, y_true


def trainModel(train_X, train_Y, test_X, test_Y,win,batch_size,start=0,models=None):
    model = Sequential()
    if models is None:
        model.add(LSTM(unit, input_shape=(train_X.shape[1], train_X.shape[2]), return_sequences=False))
        model.add(Dropout(0.1))
        model.add(Dense(train_Y.shape[1]))
        model.add(Activation("relu"))
        adam = adam_v2.Adam(learning_rate=0.01)
        model.compile(loss='mse', optimizer=adam, metrics=['acc'])
    else:
        model = load_model(models)
    log = CSVLogger(f"./log/wuhan_1_6000/{file}_{batch_size}_{unit}_{win}.csv", separator=",", append=True)
    reduce = ReduceLROnPlateau(monitor='val_acc', factor=0.5, patience=1, verbose=1,
                               mode='auto', min_delta=0.001, cooldown=0, min_lr=0.001)

    for i in range(int(start / 200),int(epoch / 200)):
        model.fit(train_X, train_Y, epochs=200, batch_size=batch_size, verbose=1, validation_split=0.1,
                  callbacks=[log, reduce])
        loss, acc = model.evaluate(test_X, test_Y, verbose=1)
        print('Loss : {}, Accuracy: {}'.format(loss, acc * 100))
        model.save(f"./model/wuhan_1_6000/{batch_size}_{unit}_{win}_{(i + 1) * 200}.h5")
    # 打印神经网络结构，统计参数数目
    model.summary()

    return model


# 计算两个经纬度之间的直线距离
def hav(theta):
    s = np.sin(theta / 2)
    return s * s


def get_distance_hav(lat0, lng0, lat1, lng1):
    EARTH_RADIUS = 6371
    # "用haversine公式计算球面两点间的距离。"
    # 经纬度转换成弧度
    lat0 = np.radians(lat0)
    lat1 = np.radians(lat1)
    lng0 = np.radians(lng0)
    lng1 = np.radians(lng1)

    dlng = np.fabs(lng0 - lng1)
    dlat = np.fabs(lat0 - lat1)
    h = hav(dlat) + np.cos(lat0) * np.cos(lat1) * hav(dlng)
    distance = 2 * EARTH_RADIUS * np.arcsin(np.sqrt(h))
    return distance


file = "wuhan_1"
#               0,90          90,270      180,359      1,180      90,270     90,180
# csv_path = ['wuhan_ping', 'wuhan_90', "wuhan_103", "wuhan_16", "wuhan_64", "wuhan_1", "wuhan_h39", "wuhan_h50"]
# 读取所有点位数据
# ais_point = pd.read_csv(f"../../tf/data/csv/{file}.csv", index_col=0)
# ais_point['updateDateFormat'] = pd.to_datetime(ais_point['updateDateFormat'])
# points_id = list(set(ais_point['traj_id']))
# points_len = len(points_id)
# train_points = ais_point.loc[ais_point.traj_id.isin(points_id[:int(points_len * 0.8)])]
# test_points = ais_point.loc[ais_point.traj_id.isin(points_id[int(points_len * 0.8):])]

train_points = pd.read_csv("../../tf/data/csv/wuhan_1_train.csv", index_col=0)
test_points = pd.read_csv("../../tf/data/csv/wuhan_1_test.csv", index_col=0)

windows = [12]
epoch = 6000
batch_sizes = [64]
units = [108]
angle1 = 0
angle2 = 360
print("file", file)
print("windows", windows)
print("units", units)
print("batch_sizes", batch_sizes)
# 生成训练数据
for wins in windows:
    train_seq, train_label, max_min = createSequence(train_points, wins, None, angle1, angle2)
    print("train_points ", train_points.shape)
    print("train_seq ", train_seq.shape)
    print("train_label ", train_label.shape)
    test_seq, test_label, _ = createSequence(test_points, wins, max_min, angle1, angle2)
    for unit in units:
        for batch in batch_sizes:
            # 训练模型
            # model = trainModel(train_seq, train_label, test_seq, test_label,wins,batch,80,"./model/wuhan_90/16_188_8_80.h5")
            model = trainModel(train_seq, train_label, test_seq, test_label,wins,batch)

            test_points_ids = list(set(test_points['traj_id']))
            y_pre = []
            y_true = []
            for ids in test_points_ids:
                test_seq, test_label, _ = createSequence(test_points.loc[test_points.traj_id == ids], wins, max_min, angle1, angle2)
                if test_seq.shape[0] <= 10:
                    continue
                print(ids)
                y_true = test_label
                for i in range(len(test_seq)):
                    y_hat = model.predict(test_seq[i].reshape(1, wins, 4))
                    y_pre.append(y_hat[0])
                y_pre = np.array(y_pre, dtype='float64')
                break
            f_y_pre, f_y_true = FNormalizeMult(y_pre, y_true, max_min)

            print(f"最大值: {max(f_y_pre[:, 4])}\n最小值: {min(f_y_pre[:, 4])}\n均值: {np.mean(f_y_pre[:, 4])}\n"
                  f"方差: {np.var(f_y_pre[:, 4])}\n标准差: {np.std(f_y_pre[:, 4])}\n中位数: {np.median(f_y_pre[:, 4])}")

            # 画测试样本数据库
            plt.subplots(figsize=(10, 8))
            plt.plot(f_y_pre[:, 0], f_y_pre[:, 1], "bo", label='预测值')
            plt.plot(f_y_true[:, 0], f_y_true[:, 1], "ro", label='真实值')
            plt.legend(loc='upper left', fontsize=15)
            plt.grid()
            plt.show()

            plt.subplots(figsize=(10, 8))
            plt.plot(f_y_pre[:, 2], "b-", label='sog,预测')
            plt.plot(f_y_true[:, 2], "r-", label='sog,真实')
            plt.legend(loc='upper left', fontsize=15)
            plt.grid()
            plt.show()

            plt.subplots(figsize=(10, 8))
            plt.plot(f_y_pre[:, 3], "b-", label='cog,预测')
            plt.plot(f_y_true[:, 3], "r-", label='cog,真实')
            plt.legend(loc='upper left', fontsize=15)
            plt.grid()
            plt.show()
exit(0)